"""This module contains the RestAdapter class, which is used to make requests to the Flowery API."""

import email.utils
import time
from asyncio import sleep as asleep
from collections.abc import Mapping
from json import JSONDecodeError
from logging import Logger
from types import SimpleNamespace

from aiohttp import (
    ClientSession,
    ClientTimeout,
    ContentTypeError,
    TraceConfig,
    TraceRequestEndParams,
    TraceRequestExceptionParams,
    TraceRequestRedirectParams,
    TraceRequestStartParams,
)
from yarl import URL

from pyflowery.exceptions import (
    ClientError,
    InternalServerError,
    ResponseError,
    RetryableException,
    RetryLimitExceeded,
    TooManyRequests,
)
from pyflowery.models import FloweryAPIConfig, Result

__all__ = ["RestAdapter"]


def _sanitize_mapping(mapping: Mapping[str, str | float | int | bool]) -> dict[str, str]:
    """Sanitize a mapping by converting all values to strings.

    Args:
        mapping (Mapping[str, str | float | int | bool]): The mapping to sanitize.

    Returns:
        The sanitized dictionary.
    """
    return {k: str(v) for k, v in mapping.items()}


class RestAdapter:
    """Underlying class for making HTTP requests.

    Args:
        config (models.FloweryAPIConfig): Configuration object for the FloweryAPI class.
    """

    def __init__(self, config: FloweryAPIConfig) -> None:
        self.config: FloweryAPIConfig = config
        self._logger: Logger = self.config.logger.getChild("rest_adapter")
        self.session: ClientSession | None = None

    async def start(self) -> ClientSession:
        headers = {"User-Agent": self.config.prepended_user_agent}
        if self.config.token:
            headers["Authorization"] = f"Bearer {self.config.token}"

        if not self.config.base_url.endswith("/"):
            self.config.base_url += "/"

        self.session = ClientSession(
            base_url=self.config.base_url,
            headers=headers,
            trace_configs=[self._trace_config()],
        )
        return self.session

    async def close(self) -> None:
        if self.session and not self.session.closed:
            await self.session.close()

    def _trace_config(self) -> TraceConfig:
        trace_config = TraceConfig()
        trace_config.on_request_start.append(self._on_request_start)
        trace_config.on_request_redirect.append(self._on_request_redirect)
        trace_config.on_request_end.append(self._on_request_end)
        trace_config.on_request_exception.append(self._on_request_exception)
        return trace_config

    @staticmethod
    def _trace_get_duration(end_time: float, trace_config_ctx: SimpleNamespace) -> str:
        start_time = getattr(trace_config_ctx, "start_time", None)
        if isinstance(start_time, float):
            return f"request took approximately {end_time - start_time:.4f}s"
        return "failed to get request duration"

    @staticmethod
    def _trace_get_urls(end_url: URL, trace_config_ctx: SimpleNamespace) -> str:
        urls = getattr(trace_config_ctx, "urls", [])
        if not isinstance(urls, list) or len(urls) <= 1:  # pyright: ignore[reportUnknownArgumentType]
            return f"'{end_url}'"

        redirect_chain = [f"'{url}'" for url in urls[:-1]]  # pyright: ignore[reportUnknownVariableType]

        full_chain = redirect_chain + [f"'{end_url}'"]

        return " >> ".join(full_chain)

    @staticmethod
    async def _on_request_start(
        session: ClientSession,  # pyright: ignore[reportUnusedParameter]
        trace_config_ctx: SimpleNamespace,
        params: TraceRequestStartParams,
    ) -> None:
        trace_config_ctx.start_time = time.monotonic()
        trace_config_ctx.urls = [params.url]

    @staticmethod
    async def _on_request_redirect(
        session: ClientSession,  # pyright: ignore[reportUnusedParameter]
        trace_config_ctx: SimpleNamespace,
        params: TraceRequestRedirectParams,
    ) -> None:
        redirect_target = params.response.headers.get("Location", None)
        if redirect_target:
            redirect_target = URL(redirect_target)
            urls = getattr(trace_config_ctx, "urls", None)
            if not isinstance(urls, list):
                trace_config_ctx.urls = [params.url, redirect_target]
                return
            urls.append(redirect_target)  # pyright: ignore[reportUnknownMemberType]
            trace_config_ctx.urls = urls

    async def _on_request_end(
        self,
        session: ClientSession,  # pyright: ignore[reportUnusedParameter]
        trace_config_ctx: SimpleNamespace,
        params: TraceRequestEndParams,
    ) -> None:
        end_time = time.monotonic()
        url = self._trace_get_urls(params.url, trace_config_ctx)

        message: str | None = None
        if params.response.status >= 400:
            try:
                data = await params.response.json()  # pyright: ignore[reportAny]
                message = str(data["error"])  # pyright: ignore[reportAny]
            except Exception:
                pass

        self._logger.debug(
            "%s %s -> %s %s%s (%s)",
            params.method,
            url,
            params.response.status,
            params.response.reason,
            f": {message}" if message else "",
            self._trace_get_duration(end_time, trace_config_ctx),
        )

    async def _on_request_exception(
        self,
        session: ClientSession,  # pyright: ignore[reportUnusedParameter]
        trace_config_ctx: SimpleNamespace,
        params: TraceRequestExceptionParams,
    ) -> None:
        end_time = time.monotonic()
        url = self._trace_get_urls(params.url, trace_config_ctx)
        self._logger.error(
            "%s request to %s encountered an exception (%s)",
            params.method,
            url,
            self._trace_get_duration(end_time, trace_config_ctx),
            exc_info=params.exception,
        )

    async def _do(
        self,
        http_method: str,
        endpoint: str,
        data: str | None = None,
        headers: Mapping[str, str | float | int | bool] | None = None,
        params: Mapping[str, str | float | int | bool] | None = None,
        timeout: float = 60,
    ) -> Result:
        """Internal method to make a request to the Flowery API. You shouldn't use this directly.

        If you need to use this method because an endpoint is missing, please open an issue on the [CoastalCommits repository](https://c.csw.im/cswimr/PyFlowery/issues).

        Args:
            http_method (str): The [HTTP method](https://developer.mozilla.org/en-US/docs/Web/HTTP/Methods) to use.
            endpoint (str): The endpoint to make the request to.
            data (str | None): The request body.
            headers (Mapping[str, str | float | int | bool] | None): Headers to send with the request. Note that `User-Agent` and `Authorization` headers will be overridden by the library if provided.
            params (Mapping[str, str | float | int | bool] | None): Query parameters to send with the request.
            timeout (float): Number of seconds to wait for the request to complete.

        Raises:
            exceptions.TooManyRequests: Raised when the Flowery API returns a 429 status code
            exceptions.ClientError: Raised when the Flowery API returns a 4xx status code
            exceptions.InternalServerError: Raised when the Flowery API returns a 5xx status code
            exceptions.RetryLimitExceeded: Raised when the retry limit defined in the [`FloweryAPIConfig`][models.FloweryAPIConfig] class (default 3) is exceeded

        Returns:
            Result: A Result object containing the status code, message, and data from the request.
        """
        if not self.session:
            self.session = await self.start()
        elif self.session and self.session.loop.is_closed():
            try:
                _ = await self.session.close()
            except Exception:
                pass
            self.session = await self.start()

        full_url = self.config.base_url + endpoint

        sanitized_headers = _sanitize_mapping(headers) if headers else {}
        sanitized_headers.update(
            {
                "User-Agent": self.config.prepended_user_agent,
            }
        )

        sanitized_params = _sanitize_mapping(params) if params else {}
        retry_counter = 0

        while retry_counter <= self.config.retry_limit:
            async with self.session.request(
                method=http_method,
                url=full_url,
                data=data,
                params=sanitized_params,
                headers=sanitized_headers,
                timeout=ClientTimeout(timeout),
            ) as response:
                try:
                    _data = await response.json()  # pyright: ignore[reportAny]
                except (JSONDecodeError, ContentTypeError):
                    _data = await response.read()

                cache_hit = None
                if raw_cache_hit := response.headers.get("X-Cache-Status", None):
                    cache_hit = True if raw_cache_hit == "HIT" else False

                cache_expires = None
                if raw_cache_expires := response.headers.get("Expires", None):
                    cache_expires = email.utils.parsedate_to_datetime(raw_cache_expires)

                result = Result(
                    success=response.status <= 400,
                    status_code=response.status,
                    message=response.reason or "Unknown error",
                    data=_data,
                    cache_hit=cache_hit,
                    cache_expires=cache_expires,
                    raw_response=response,
                )

                try:
                    if result.status_code == 429:
                        raise TooManyRequests(str(result), result)
                    if 400 <= result.status_code < 500:
                        raise ClientError(str(result), result)
                    if 500 <= result.status_code < 600:
                        raise InternalServerError(str(result), result)
                except RetryableException as e:
                    if self.config.retry_limit <= 0:
                        raise e
                    if retry_counter < self.config.retry_limit:
                        interval = self.config.interval * retry_counter
                        self._logger.exception("%s - retrying in %s seconds", e, interval)
                        retry_counter += 1
                        await asleep(interval)
                        continue
                    raise RetryLimitExceeded(
                        message=f"Request failed more than {self.config.retry_limit} times, not retrying.", result=result
                    ) from e
                return result
        # this shouldn't ever happen but is here to keep typecheckers happy
        message = "Empty Response! Something went wrong in RestAdapter!"
        raise ResponseError(
            message,
            Result(success=False, status_code=400, message=message),
        )

    async def get(
        self,
        endpoint: str,
        data: str | None = None,
        headers: Mapping[str, str | float | int | bool] | None = None,
        params: Mapping[str, str | float | int | bool] | None = None,
        timeout: float = 60,
    ) -> Result:
        """Make a GET request to the Flowery API. You should almost never have to use this directly.

        If you need to use this method because an endpoint is missing, please open an issue on the [CoastalCommits repository](https://c.csw.im/cswimr/PyFlowery/issues).

        Args:
            endpoint (str): The endpoint to make the request to.
            data (str | None): The request body.
            headers (Mapping[str, str | float | int | bool] | None): Headers to send with the request. Note that `User-Agent` and `Authorization` headers will be overridden by the library if provided.
            params (Mapping[str, str | float | int | bool] | None): Query parameters to send with the request.
            timeout (float): Number of seconds to wait for the request to complete.

        Raises:
            exceptions.TooManyRequests: Raised when the Flowery API returns a 429 status code
            exceptions.ClientError: Raised when the Flowery API returns a 4xx status code
            exceptions.InternalServerError: Raised when the Flowery API returns a 5xx status code
            exceptions.RetryLimitExceeded: Raised when the retry limit defined in the [`FloweryAPIConfig`][models.FloweryAPIConfig] class (default 3) is exceeded

        Returns:
            An object containing the status code, message, and data from the request.
        """
        return await self._do(http_method="GET", endpoint=endpoint, data=data, headers=headers, params=params, timeout=timeout)

    async def post(
        self,
        endpoint: str,
        data: str | None = None,
        headers: Mapping[str, str | float | int | bool] | None = None,
        params: Mapping[str, str | float | int | bool] | None = None,
        timeout: float = 60,
    ) -> Result:
        """Make a POST request to the Flowery API. You should almost never have to use this directly.

        If you need to use this method because an endpoint is missing, please open an issue on the [CoastalCommits repository](https://c.csw.im/cswimr/PyFlowery/issues).

        Args:
            endpoint (str): The endpoint to make the request to.
            data (str | None): The request body.
            headers (Mapping[str, str | float | int | bool] | None): Headers to send with the request. Note that `User-Agent` and `Authorization` headers will be overridden by the library if provided.
            params (Mapping[str, str | float | int | bool] | None): Query parameters to send with the request.
            timeout (float): Number of seconds to wait for the request to complete.

        Raises:
            exceptions.TooManyRequests: Raised when the Flowery API returns a 429 status code
            exceptions.ClientError: Raised when the Flowery API returns a 4xx status code
            exceptions.InternalServerError: Raised when the Flowery API returns a 5xx status code
            exceptions.RetryLimitExceeded: Raised when the retry limit defined in the [`FloweryAPIConfig`][models.FloweryAPIConfig] class (default 3) is exceeded

        Returns:
            An object containing the status code, message, and data from the request.
        """
        return await self._do(http_method="POST", endpoint=endpoint, data=data, headers=headers, params=params, timeout=timeout)



from pyflowery.models import Result

__all__ = [
    "HttpException",
    "RetryableException",
    "ResponseError",
    "InternalServerError",
    "ClientError",
    "TooManyRequests",
    "RetryLimitExceeded",
]


class HttpException(Exception):
    def __init__(self, message: str, result: Result) -> None:
        super().__init__(message)
        self.message: str = message
        self.result: Result = result


class RetryableException(HttpException):
    """Class for exceptions which should be retryable. Never raise this yourself."""


class ResponseError(HttpException):
    """Raised when an API response is empty or has an unexpected format."""

    def __init__(self, message: str, result: Result) -> None:
        super().__init__(message, result)
        self.message: str = "Invalid response from Flowery API: " + message


class InternalServerError(RetryableException):
    """Raised when the API returns a 5xx status code."""


class ClientError(HttpException):
    """Raised when the API returns a 4xx status code."""


class TooManyRequests(RetryableException):
    """Raised when the API returns a 429 status code."""

    def __init__(self, message: str, result: Result) -> None:
        super().__init__(message, result)


class RetryLimitExceeded(HttpException):
    """Raised when the retry limit is exceeded."""



from pyflowery import FloweryAPI
from pyflowery.utils import call_async

from .config import config

api = FloweryAPI(config)


def test_cache_population():
    assert len(api._voices_cache) == 0  # pyright: ignore[reportPrivateUsage]
    _ = call_async(api.fetch_voices(), api.logger)
    assert len(api._voices_cache) > 0  # pyright: ignore[reportPrivateUsage]
    _ = call_async(api.close(), api.logger)



import asyncio
import threading
from collections.abc import Coroutine
from logging import Logger
from typing import TypeVar

__all__ = ["call_async"]

R = TypeVar("R")


def call_async(
    coroutine: Coroutine[None, None, R],
    logger: Logger,
) -> asyncio.Task[R] | R:
    """
    Call an async function safely in both synchronous and asynchronous contexts.

    - If no event loop is running in this thread: runs the coroutine with asyncio.run and returns the result.
    - If an event loop is running in this thread: runs the coroutine in a new thread with its own loop and blocks
      until completion, returning the result.
    - Returns an asyncio.Task if you explicitly want to schedule in the current loop without blocking.

    Args:
        coroutine: the coroutine to await
        logger: standard library Logger for reporting what’s happening

    Returns:
        The completed result (R) or an asyncio.Task[R] if scheduled in a live loop without blocking.
    """
    name = coroutine.__qualname__

    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None

    if loop and loop.is_running():
        # There's a running loop; run the coroutine in a separate thread to avoid deadlock
        logger.info("Event loop running; offloading %r to a background thread", name)

        result_container: dict[str, R] = {}

        def target():
            result_container["result"] = asyncio.run(coroutine)

        t = threading.Thread(target=target)
        t.start()
        t.join()  # Block until completion
        return result_container["result"]
    else:
        # No loop running; safe to run normally
        logger.info("No running event loop; running %r via asyncio.run", name)
        return asyncio.run(coroutine)



import sys
from datetime import datetime
from enum import StrEnum
from logging import INFO, Formatter, Logger, StreamHandler, getLogger
from typing import Any, ClassVar, Self

import aiohttp
from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import override

from pyflowery.version import version

__all__ = ["Result", "AudioFormat", "Voice", "Language", "TTSResponse", "FloweryAPIConfig"]


class Result(BaseModel):
    """Result returned from low-level RestAdapter."""

    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    success: bool
    """Whether or not the request was successful."""
    status_code: int
    """The [HTTP status code](https://developer.mozilla.org/en-US/docs/Web/HTTP/Reference/Status) returned by the request."""
    message: str = ""
    """Human readable result message from the request."""
    data: list[dict[str, Any]] | dict[str, Any] | bytes = {}  # pyright: ignore[reportExplicitAny]
    """The data returned by the request."""
    cache_hit: bool | None = None
    """Whether or not the data in this result was returned via a cache from the external API.

    Will be `None` in cases where the external API does not tell us if the result was cached.
    """
    cache_expires: datetime | None = None
    """The timestamp at which the data in this result will expire, if it was cached by the external API.

    Will be `None` in cases where the external API does not tell us if the result was cached.
    """
    raw_response: aiohttp.ClientResponse | None = None
    """The raw response object from the request."""

    @override
    def __str__(self) -> str:
        message = self.message
        if not self.success:
            if isinstance(self.data, dict) and (error := self.data.get("error")):
                message = f"{message}: {error}"
        if self.cache_hit is not None:
            return f"{self.status_code} {message} (Cache {'hit' if self.cache_hit else 'miss'})"
        return f"{self.status_code} {message}"


class AudioFormat(StrEnum):
    """List of audio formats accepted by Flowery."""

    MP3 = "audio/mpeg"
    """[MP3](https://en.wikipedia.org/wiki/MP3) audio format."""
    OPUS = "audio/ogg; codecs=opus"
    """[OGG](https://en.wikipedia.org/wiki/Ogg) [Opus](https://en.wikipedia.org/wiki/Opus_(audio_format)) audio format."""
    VORBIS = "audio/ogg; codecs=vorbis"
    """[OGG](https://en.wikipedia.org/wiki/Ogg) [Vorbis](https://en.wikipedia.org/wiki/Vorbis) audio format."""
    AAC = "audio/x-aac"
    """[AAC](https://en.wikipedia.org/wiki/Advanced_Audio_Coding) audio format."""
    WAV = "audio/wav"
    """[WAV](https://en.wikipedia.org/wiki/WAV) audio format."""
    FLAC = "audio/x-flac"
    """[FLAC](https://en.wikipedia.org/wiki/FLAC) audio format."""


class Language(BaseModel):
    """Language object returned from the Flowery API."""

    name: str | None = None
    """The name associated with the language."""
    code: str
    """The code associated with the language."""

    @override
    def __eq__(self, other: object) -> bool:
        if isinstance(other, Language):
            return True if self.code == other.code else False
        return False

    @classmethod
    def english(cls) -> list[Self]:
        return [
            cls(code="en-US", name="English (United States)"),
            cls(code="en_US", name="English (United States)"),
            cls(code="en-GB", name="English (Great Britain)"),
            cls(code="en_GB", name="English (Great Britain)"),
            cls(code="en", name="English"),
        ]


class Voice(BaseModel):
    """Voice object returned from the Flowery API."""

    id: str
    """UUID of the voice."""
    name: str
    """Name of the voice."""
    gender: str
    """Gender of the voice."""
    source: str
    """Which provider the voice comes from."""
    language: Language
    """Which language the voice is meant to be used for."""


class TTSResponse(BaseModel):
    """Object containing data from a TTS query."""

    data: bytes
    """The synthesized TTS data in bytes."""
    text: str
    """The SSML text used to generate the TTS data."""
    voice: Voice | None = None
    """Whether the text was translated into the voice's set language before being synthesized."""
    audio_format: AudioFormat
    """What format the data is in."""
    result: Result
    """The internal result object for use in debugging."""


def _create_logger(level: str | int = INFO) -> Logger:
    """Creates a logger for the library."""
    logger = getLogger("pyflowery")
    if logger.handlers:
        return logger

    logger.setLevel(level)
    logger.propagate = False

    handler = StreamHandler(sys.stderr)
    handler.setLevel(level)

    formatter = Formatter(fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    handler.setFormatter(formatter)

    logger.addHandler(handler)
    return logger


class FloweryAPIConfig(BaseModel):
    """Configuration used when making requests to the Flowery API.

    Example:
        ```python
        from pyflowery import FloweryAPI, FloweryAPIConfig

        config = FloweryAPIConfig(
            user_agent = "PyFloweryDocumentation/1.0.0",
            token = "hello world",
        )

        api = FloweryAPI(config=config)
        ```
    """

    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True, validate_default=True)

    user_agent: str
    """User-Agent string to use in HTTP requests. Required as of 2.1.0."""
    token: str | None = None
    """Authorization token to use in HTTP requests. Authorization is only required if you want increased ratelimits.
            See the [Flowery API documentation](https://flowery.pw/docs) for more details."""
    base_url: str = "https://api.flowery.pw/v2"
    """The base url for the Flowery API."""
    logger: Logger = Field(default_factory=_create_logger)
    """Logger to use for logging messages. One will be automatically created if not provided here."""
    retry_limit: int = 3
    """Number of times to retry a request before giving up."""
    interval: int = 5
    """Seconds to wait between each retried request, multiplied by how many attempted requests have been made."""

    @property
    def prepended_user_agent(self) -> str:
        """Return the user_agent with the PyFlowery module version prepended."""
        return f"PyFlowery/{version} {self.user_agent} (Python {sys.version})"



import asyncio
from collections.abc import Mapping
from copy import deepcopy
from logging import Logger
from typing import overload

from lxml import etree
from pydantic import ValidationError

from pyflowery import exceptions, models, rest_adapter, utils

__all__ = ["FloweryAPI"]


class FloweryAPI:
    """Main class for interacting with the Flowery API.

    Example:
        ```python
        from pyflowery import FloweryAPI, FloweryAPIConfig

        config = FloweryAPIConfig(
            user_agent = "PyFloweryDocumentation/1.0.0",
            token = "hello world",
            allow_truncation = False,
        )

        api = FloweryAPI(config=config)
        ```
    """

    def __init__(self, config: models.FloweryAPIConfig) -> None:
        self.config: models.FloweryAPIConfig = config
        """Object for configuring requests to the Flowery API."""
        self.adapter: rest_adapter.RestAdapter = rest_adapter.RestAdapter(config=config)
        """Rest Adapter used for making HTTP requests."""
        self.logger: Logger = self.config.logger
        """The logger used internally."""

        self._voices_cache: Mapping[str, models.Voice] = {}
        self._voices_populating: asyncio.Task[Mapping[str, models.Voice]] | None = None

    async def close(self) -> None:
        """Close the currently open client session."""
        await self.adapter.close()

    def _ensure_voices_cache(self) -> None:
        if self._voices_cache:
            return
        _ = utils.call_async(self.fetch_voices(), self.logger)

    def get_voice(self, voice_id: str) -> models.Voice | None:
        """Get a voice from the cache using its ID.

        Example:
            ```python
            from pyflowery import FloweryAPI, FloweryAPIConfig

            api = FloweryAPI(config=FloweryAPIConfig(user_agent="PyFloweryDocumentation/1.0.0"))
            voice = api.get_voice("372a5e97-1645-563a-9097-36bd83184ab4")
            # Voice(
            #   id='372a5e97-1645-563a-9097-36bd83184ab4',
            #   name='Xiaoyi', gender='Female', source='Microsoft Azure',
            #   language=Language(name='Chinese (China)', code='zh-CN')
            # )
            ```

        Args:
            voice_id (str): The ID of the voice to retrieve from the cache.

        Returns:
            The matching [`Voice`][models.Voice] if found, otherwise `None`.
        """
        self._ensure_voices_cache()
        return self._voices_cache.get(voice_id.lower())

    @overload
    def get_voices(self) -> tuple[models.Voice, ...]: ...

    @overload
    def get_voices(
        self,
        *,
        name: str | None = None,
        gender: str | None = None,
        source: str | None = None,
        languages: list[models.Language] | models.Language | None = None,
    ) -> tuple[models.Voice, ...] | None: ...

    def get_voices(
        self,
        *,
        name: str | None = None,
        gender: str | None = None,
        source: str | None = None,
        languages: list[models.Language] | models.Language | None = None,
    ) -> tuple[models.Voice, ...] | None:
        """Get a filtered set of voices from the cache.

        Example:
            ```python
            from pyflowery import FloweryAPI, FloweryAPIConfig, Language

            api = FloweryAPI(config=FloweryAPIConfig(user_agent="PyFloweryDocumentation/1.0.0"))
            voices = api.get_voices(source="TikTok", languages=Language(name="English (United States)", code="en-US"))
            # (
            #   Voice(
            #       id='fa3ea565-121f-5efd-b4e9-59895c77df23',
            #       name='Alexander', gender='Male', source='TikTok',
            #       language=Language(name='English (United States)', code='en-US')
            #   ),
            #   Voice(
            #       id='a55b0ad0-84c8-597d-832b-0bc4c8777054',
            #       name='Alto', gender='Female', source='TikTok',
            #       language=Language(name='English (United States)', code='en-US')
            #   ), ...
            ```

        Args:
            name: The name to filter results by.
            gender: The gender to filter results by.
            source: The source to filter results by.
            languages: The languages to filter results by.

        Returns:
            (tuple[models.Voice, ...]): All voices when no filters are given.
            (tuple[models.Voice, ...] | None): Filtered voices (or None if no matches).
        """
        self._ensure_voices_cache()

        if not any([name, gender, source, languages]):
            return tuple(self._voices_cache.values())

        if languages and not isinstance(languages, list):
            languages = [languages]

        language_codes = {lang.code for lang in languages} if languages else None

        voices: list[models.Voice] = []

        for voice in self._voices_cache.values():
            if name is not None and voice.name.lower() != name.lower():
                continue
            if gender is not None and voice.gender.lower() != gender.lower():
                continue
            if source is not None and voice.source.lower() != source.lower():
                continue
            if language_codes is not None and voice.language.code not in language_codes:
                continue
            voices.append(voice)

        return tuple(voices) if len(voices) > 0 else None

    async def fetch_voices(self) -> Mapping[str, models.Voice]:
        """Fetch a mapping of voice IDs to voices from the Flowery API.

        Calling this method will repopulate the built-in voices cache.

        Raises:
            exceptions.TooManyRequests: Raised when the Flowery API returns a 429 status code.
            exceptions.ClientError: Raised when the Flowery API returns a 4xx status code.
            exceptions.InternalServerError: Raised when the Flowery API returns a 5xx status code.
            exceptions.ResponseError: Raised when the Flowery API returns an empty response or a response with an unexpected format.
            exceptions.RetryLimitExceeded: Raised when the retry limit defined in the `FloweryAPIConfig` class (default 3) is exceeded.

        Returns:
            A mapping of voice IDs to voices.
        """
        request = await self.adapter.get(endpoint="voices")
        if request.success is not True or not isinstance(request.data, dict):
            raise exceptions.ResponseError(f"Invalid response from Flowery API: {request.data!r}", request)

        data: list[dict[str, str]] = request.data.get("voices", [])  # pyright: ignore[reportAny]
        voices: dict[str, models.Voice] = {}
        for voice in data:
            try:
                v = models.Voice.model_validate(voice)
                voices[v.id.lower()] = v
            except ValidationError as e:
                self.config.logger.exception("Failed to validate voice data. Voice Data: %s, Error: %s", voice, str(e))
                continue
        self._voices_cache = deepcopy(voices)
        self.logger.debug("Voices cache populated with %d entries!", len(self._voices_cache))
        return voices

    async def fetch_tts(
        self,
        text: str | etree._Element,  # pyright: ignore[reportPrivateUsage]
        voice: models.Voice,
        audio_format: models.AudioFormat = models.AudioFormat.MP3,
        speed: float = 1.0,
    ) -> models.TTSResponse:
        """Fetch a TTS audio file from the Flowery API.

        Args:
            text (str | lxml.etree._Element): The SSML text to convert to speech. This will be contained within the body of a `<voice>` tag.
                See the [Flowery SSML documentation](https://flowery.pw/docs/ssml) for more information on SSML.
            voice (models.Voice): The voice to use for the speech.
            audio_format (models.AudioFormat): The audio format to return.

        Raises:
            ValueError: Raised when the provided text is too long.
            exceptions.TooManyRequests: Raised when the Flowery API returns a 429 status code.
            exceptions.ClientError: Raised when the Flowery API returns a 4xx status code.
            exceptions.InternalServerError: Raised when the Flowery API returns a 5xx status code.
            exceptions.ResponseError: Raised when the Flowery API returns an empty response or a response with an unexpected format.
            exceptions.RetryLimitExceeded: Raised when the retry limit defined in the [`FloweryAPIConfig`][models.FloweryAPIConfig] class (default 3) is exceeded.

        Returns:
            An object containing the parameters used to synthesize the text, as well as the raw tts data in bytes.
        """
        if len(text) > 4096:
            raise ValueError("Text must be less than or equal to 4096 characters")

        headers = {"Content-Type": "application/xml; charset=UTF-8", "Accept": audio_format}

        speak = etree.Element("speak")
        ssml_voice = etree.SubElement(speak, "voice", name=voice.id)
        if isinstance(text, str):
            try:
                ssml_voice.append(etree.fromstring(text))
            except etree.XMLSyntaxError:
                ssml_voice.text = text
        else:
            ssml_voice.append(text)

        body = etree.tostring(speak, pretty_print=True).decode()

        params = {
            "text": text,
            "audio_format": audio_format,
            "speed": speed,
        }
        if voice:
            params["voice"] = voice.id

        request = await self.adapter.post(endpoint="tts", headers=headers, data=body, timeout=180)

        if isinstance(request.data, bytes):
            return models.TTSResponse(data=request.data, text=body, voice=voice, audio_format=audio_format, result=request)
        raise exceptions.ResponseError(f"Invalid response from Flowery API: {request.data!r}", request)



# This file is automatically generated on build, do not edit!
# Do not commit this file to source control!

__all__ = [
    "__version__",
    "__version_tuple__",
    "version",
    "version_tuple"
]

version: str
__version__: str
__version_tuple__: tuple[int | str, ...]
version_tuple: tuple[int | str, ...]

__version__ = version = '4.0.0'
__version_tuple__ = version_tuple = (4, 0, 0)




from datetime import timedelta
from random import choice
from typing import TypeAlias

import pytest
from pydantic import ValidationError

from pyflowery import AudioFormat, FloweryAPI, Result, TTSResponse, Voice
from pyflowery.utils import call_async

from ..config import config

api = FloweryAPI(config)


class TestTTSResponse:
    # Sample data for testing
    sample_voice: Voice = choice(api.get_voices())
    sample_audio_format: AudioFormat = choice(tuple(AudioFormat))

    T: TypeAlias = dict[str, bytes | str | Voice | bool | timedelta | AudioFormat | Result | float]

    @pytest.fixture
    def valid_data(self) -> T:
        return {
            "data": b"audio_data",
            "text": "Hello world",
            "voice": self.sample_voice,
            "audio_format": self.sample_audio_format,
            "result": Result(success=True, status_code=200),
        }

    def test_valid_model_creation(self, valid_data: T):
        """Test that a model can be created with valid data"""
        response = TTSResponse.model_validate(valid_data)

        assert response.data == valid_data["data"]
        assert response.text == valid_data["text"]
        assert response.voice == valid_data["voice"]
        assert response.audio_format == valid_data["audio_format"]
        assert response.result == valid_data["result"]

    def test_required_fields(self, valid_data: T):
        """Test that required fields are enforced"""
        required_fields = ["data", "text", "audio_format", "result"]

        for field in required_fields:
            data = valid_data.copy()
            _ = data.pop(field)
            with pytest.raises(ValidationError):
                _ = TTSResponse.model_validate(data)

    def test_default_values(self, valid_data: T):
        """Test default values are set correctly"""
        data = {"data": b"audio", "text": "text", "audio_format": self.sample_audio_format, "result": valid_data["result"]}

        response = TTSResponse.model_validate(data)

        assert response.voice is None


_ = call_async(api.close(), config.logger)



# pyright: reportImportCycles=false
from pyflowery.exceptions import ClientError, InternalServerError, ResponseError, RetryLimitExceeded, TooManyRequests
from pyflowery.models import AudioFormat, FloweryAPIConfig, Language, Result, TTSResponse, Voice
from pyflowery.pyflowery import FloweryAPI
from pyflowery.version import __version__

__all__ = [
    "FloweryAPI",
    "FloweryAPIConfig",
    "AudioFormat",
    "Language",
    "Result",
    "TTSResponse",
    "Voice",
    "__version__",
    "ResponseError",
    "ClientError",
    "InternalServerError",
    "RetryLimitExceeded",
    "TooManyRequests",
]
